@@ -1,4 +1,16 @@
 # Copyright 2012 Google Inc. All Rights Reserved.
+#
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#
+#    http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing,
+# software distributed under the License is distributed on an
+# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
+# either express or implied. See the License for the specific
+# language governing permissions and limitations under the License.
 
 """Python wrappers for the Google Storage RESTful API."""
 
@@ -11,7 +23,6 @@ __all__ = ['ReadBuffer',
           ]
 
 import collections
-import logging
 import os
 import urlparse
 
@@ -197,8 +208,8 @@ class ReadBuffer(object):
 
     self._request_next_buffer()
 
-    status, headers, _ = self._api.head_object(path)
-    errors.check_status(status, [200], path, resp_headers=headers)
+    status, headers, content = self._api.head_object(path)
+    errors.check_status(status, [200], path, resp_headers=headers, body=content)
     self._file_size = long(headers['content-length'])
     self._check_etag(headers.get('etag'))
     if self._file_size == 0:
@@ -419,7 +430,8 @@ class ReadBuffer(object):
     headers = {'Range': 'bytes=' + content_range}
     status, resp_headers, content = yield self._api.get_object_async(
         self._path, headers=headers)
-    errors.check_status(status, [200, 206], self._path, headers, resp_headers)
+    errors.check_status(status, [200, 206], self._path, headers, resp_headers,
+                        body=content)
     self._check_etag(resp_headers.get('etag'))
     raise ndb.Return(content)
 
@@ -588,7 +600,9 @@ class StreamingBuffer(object):
 
   _blocksize = 256 * 1024
 
-  _maxrequestsize = 16 * _blocksize
+  _flushsize = 8 * _blocksize
+
+  _maxrequestsize = 9 * 4 * _blocksize
 
   def __init__(self,
                api,
@@ -604,12 +618,16 @@ class StreamingBuffer(object):
         delegate to Google Cloud Storage.
       gcs_headers: additional gs headers as a str->str dict, e.g
         {'x-goog-acl': 'private', 'x-goog-meta-foo': 'foo'}.
+    Raises:
+      IOError: When this location can not be found.
     """
     assert self._maxrequestsize > self._blocksize
     assert self._maxrequestsize % self._blocksize == 0
+    assert self._maxrequestsize >= self._flushsize
 
     self._api = api
     self._path = path
+
     self.name = api_utils._unquote_filename(path)
     self.closed = False
 
@@ -623,8 +641,9 @@ class StreamingBuffer(object):
       headers['content-type'] = content_type
     if gcs_headers:
       headers.update(gcs_headers)
-    status, resp_headers, _ = self._api.post_object(path, headers=headers)
-    errors.check_status(status, [201], path, headers, resp_headers)
+    status, resp_headers, content = self._api.post_object(path, headers=headers)
+    errors.check_status(status, [201], path, headers, resp_headers,
+                        body=content)
     loc = resp_headers.get('location')
     if not loc:
       raise IOError('No location header found in 201 response')
@@ -685,18 +704,18 @@ class StreamingBuffer(object):
     self._buffer.append(data)
     self._buffered += len(data)
     self._offset += len(data)
-    if self._buffered >= self._blocksize:
+    if self._buffered >= self._flushsize:
       self._flush()
 
   def flush(self):
-    """Dummy API.
+    """Flush as much as possible to GCS.
 
-    This API is provided because the zipfile module uses it.  It is a
-    no-op because Google Storage *requires* that all writes except for
-    the final one are multiples on 256K bytes aligned on 256K-byte
-    boundaries.
+    GCS *requires* that all writes except for the final one align on
+    256KB boundaries. So the internal buffer may still have < 256KB bytes left
+    after flush.
     """
     self._check_open()
+    self._flush(finish=False)
 
   def tell(self):
     """Return the total number of bytes passed to write() so far.
@@ -725,45 +744,44 @@ class StreamingBuffer(object):
   def _flush(self, finish=False):
     """Internal API to flush.
 
-    This is called only when the total amount of buffered data is at
+    Buffer is flushed to GCS only when the total amount of buffered data is at
     least self._blocksize, or to flush the final (incomplete) block of
     the file with finish=True.
     """
-    flush_len = 0 if finish else self._blocksize
-
-    while self._buffered >= flush_len:
-      buffer = []
-      buffered = 0
+    while ((finish and self._buffered >= 0) or
+           (not finish and self._buffered >= self._blocksize)):
+      tmp_buffer = []
+      tmp_buffer_len = 0
 
+      excess = 0
       while self._buffer:
         buf = self._buffer.popleft()
         size = len(buf)
         self._buffered -= size
-        buffer.append(buf)
-        buffered += size
-        if buffered >= self._maxrequestsize:
+        tmp_buffer.append(buf)
+        tmp_buffer_len += size
+        if tmp_buffer_len >= self._maxrequestsize:
+          excess = tmp_buffer_len - self._maxrequestsize
+          break
+        if not finish and (
+            tmp_buffer_len % self._blocksize + self._buffered <
+            self._blocksize):
+          excess = tmp_buffer_len % self._blocksize
           break
 
-      if buffered > self._maxrequestsize:
-        excess = buffered - self._maxrequestsize
-      elif finish:
-        excess = 0
-      else:
-        excess = buffered % self._blocksize
-
       if excess:
-        over = buffer.pop()
+        over = tmp_buffer.pop()
         size = len(over)
         assert size >= excess
-        buffered -= size
+        tmp_buffer_len -= size
         head, tail = over[:-excess], over[-excess:]
         self._buffer.appendleft(tail)
         self._buffered += len(tail)
         if head:
-          buffer.append(head)
-          buffered += len(head)
+          tmp_buffer.append(head)
+          tmp_buffer_len += len(head)
 
-      data = ''.join(buffer)
+      data = ''.join(tmp_buffer)
       file_len = '*'
       if finish and not self._buffered:
         file_len = self._written + len(data)
@@ -792,14 +810,14 @@ class StreamingBuffer(object):
     else:
       headers['content-range'] = ('bytes */%s' % file_len)
 
-    status, response_headers, _ = self._api.put_object(
+    status, response_headers, content = self._api.put_object(
         self._path_with_token, payload=data, headers=headers)
     if file_len == '*':
       expected = 308
     else:
       expected = 200
     errors.check_status(status, [expected], self._path, headers,
-                        response_headers,
+                        response_headers, content,
                         {'upload_path': self._path_with_token})
 
   def _get_offset_from_gcs(self):
@@ -812,10 +830,10 @@ class StreamingBuffer(object):
       -1 means nothing has been written.
     """
     headers = {'content-range': 'bytes */*'}
-    status, response_headers, _ = self._api.put_object(
+    status, response_headers, content = self._api.put_object(
         self._path_with_token, headers=headers)
     errors.check_status(status, [308], self._path, headers,
-                        response_headers,
+                        response_headers, content,
                         {'upload_path': self._path_with_token})
     val = response_headers.get('range')
     if val is None:
